"
 COPYRIGHT (c) 1992 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Link subclass:#Process
	 instanceVariableNames:'id prio state startBlock name 
				restartable interruptActions 
				exitAction singleStepping
				emergencySignalHandler'
	 classVariableNames:'TerminateSignal'
	 poolDictionaries:''
	 category:'Kernel-Processes'
!

Process comment:'
COPYRIGHT (c) 1992 by Claus Gittinger
	     All Rights Reserved

$Header: /cvs/stx/stx/libbasic/Process.st,v 1.15 1994-10-10 00:27:33 claus Exp $
'!

!Process class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1992 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /cvs/stx/stx/libbasic/Process.st,v 1.15 1994-10-10 00:27:33 claus Exp $
"
!

documentation
"
    Instances of Process represent lightweight smalltalk processes 
    (i.e. threads). These all run in a shared smalltalk/X address space,
    and can thus access and communicate via any objects.
    Do not confuse these with (heavy-weight) unix processes, which are
    created differently, and do NOT run in the same address space.
    Also notice, that heavy-weight process creation takes much longer.
    (see OperatingSystemclass>>fork).

    Processes are typically created by sending #fork or #forkAt: to a block.
    Scheduling is done by Processor, which is the sole instance of 
    ProcessorScheduler.

    Notice: in Smalltalk/X, processes are gone, when an image is restarted;
    this means, that you have to take care of process re-creation yourself.
    Usually, this is done by depending on ObjectMemory, recreating the
    process(s) when the #returnFromSnapshot-change notifiaction arrives.

    A later version will allow specification of automatic restart, but
    thats not yet implemented.

    Instance variables:

	id                     <SmallInteger>   a unique process-id
	prio                   <SmallInteger>   the processes priority
	state                  <Symbol>         the processes state
						(for process-monitor)
	startBlock             <Block>          the startup-block
	name                   <String-or-nil>  the processes name (if any)
						(for process-monitor)
	restartable            <Boolean>        is restartable (not yet implemented)
	interruptActions       <Collection>     interrupt actions as defined by interruptWith:,
						performed at interrupt time
	exitAction             <Block>          action to perform on termination
	emergencySignalHandler <Block>          can be used for per-process
						emergency signal handling
    Class variables:

	TerminateSignal         <Signal>        signal used to terminate processes
						(should not be cought - or at least
						 rejected in handlers)
"
! !

!Process class methodsFor:'initialization'!

initialize
    TerminateSignal isNil ifTrue:[
	TerminateSignal := Signal new mayProceed:true.
	TerminateSignal nameClass:self message:#terminateSignal.
	TerminateSignal notifierString:'unhandled process termination'.
    ]
! !

!Process class methodsFor:'signal access'!

terminateSignal
    ^ TerminateSignal
! !

!Process class methodsFor:'instance creation'!

for:aBlock priority:aPrio
    "create a new (unscheduled) process which will execute aBlock at
     a given priority, once scheduled. The process will start execution once
     it gets a #resume-message."

    ^ self new for:aBlock priority:aPrio
! !

!Process methodsFor:'private'!

for:aBlock priority:aPrio
    "setup the new process - the receiver is not scheduled for
     execution, to get it running, send it #resume"

    |nm|

    prio := aPrio.
    startBlock := aBlock.

    (Processor newProcessFor:self) ifFalse:[ 
	"for some reason, the Processor was unable to create
	 a VM process for me ...."
	^ nil
    ].

    "
     give me a user-friendly name
    "
    (nm := Processor activeProcess name) notNil ifTrue:[
	"
	 avoid name inflation
	"
	(nm endsWith:'(sub)') ifFalse:[
	    nm := nm , ' (sub)'
	].
	name := nm
    ]
! !

!Process methodsFor:'accessing'!

state
    "return a symbol describing the processes state"

    ^ state
!

state:aSymbol
    "set the state - only to be used from scheduler"

    state := aSymbol
!

startBlock
    "return the processes startup-block"

    ^ startBlock
!

emergencySignalHandler:aOneArgBlock
    "set the emergencySignalHandler block.
     See Signal>>documentation for more info."

    emergencySignalHandler := aOneArgBlock
!

emergencySignalHandler
    "return the emergencySignalHandler block.
     See Signal>>documentation for more info."

    ^ emergencySignalHandler
!

priority
    "return the receivers priority"

    ^ prio
!

priority:aNumber
    "set my priority"

    Processor changePriority:aNumber for:self.
!

changePriority:aNumber
    "same as priority:, but returns the old priority.
     (cannot do this in priority: for ST-80 compatibility)"

    |oldPrio|

    oldPrio := prio.
    Processor changePriority:aNumber for:self.
    ^ oldPrio
!

isSingleStepping
    ^ singleStepping
!

singleStep:aBoolean
    singleStepping := aBoolean
!

id
    "return the processes id"

    ^ id
!

name
    "return the processes name"

    ^ name
!

name:aString
    "set the processes name"

    name := aString
!

nameOrId
    "return a string to identify the process - either name or id"

    name notNil ifTrue:[^ name].
    ^ id printString
!

exitAction
    "return the processes exit action"

    ^ exitAction
!

exitAction:aBlock
    "set the processes exit action"

    exitAction := aBlock
!

suspendedContext
    "return the processes suspended context"

%{  /* NOCONTEXT */
    extern OBJ __threadContext();
    OBJ i;

    if (_isSmallInteger(i = _INST(id))) {
	RETURN (__threadContext(_intVal(i)));
    }
%}
.
    ^ nil
!

usedStackSize
    "Return the processes current stack size.
     This method is for monitoring purposes only."

%{  /* NOCONTEXT */
    extern int __threadUsedStackSize();
    OBJ i;

    if (_isSmallInteger(i = _INST(id))) {
	RETURN( _MKSMALLINT(__threadUsedStackSize(_intVal(i))) );
    }
%}
.
    ^ nil
!

totalStackSize
    "return the processes maximum used stack size.
     This method is for monitoring purposes only."

%{  /* NOCONTEXT */
    extern int __threadTotalStackSize();
    OBJ i;

    if (_isSmallInteger(i = _INST(id))) {
	RETURN( _MKSMALLINT(__threadTotalStackSize(_intVal(i))) );
    }
%}
.
    ^ nil
!

maximumStackSize
    "returns the processes stack limit - i.e. the process will be 
     interrupted with a recursionSignal-raise, if it ever
     needs more stack (in bytes) than this number"

%{  /* NOCONTEXT */
    extern int __threadMaxStackSize();
    OBJ i;

    if (_isSmallInteger(i = _INST(id))) {
	RETURN( _MKSMALLINT(__threadMaxStackSize(_intVal(i))) );
    }
%}
.
    ^ nil
!
 
setMaximumStackSize:limit
    "sets the processes stack limit - i.e. the process will be
     interrupted with a recursionSignal-raise, if it ever
     needs more stack (in bytes) than this number.
     Returns the old value."

%{  /* NOCONTEXT */
    extern int __threadSetMaxStackSize();
    OBJ i;

    if (_isSmallInteger(i = _INST(id)) 
     && _isSmallInteger(limit) ) {
	RETURN ( _MKSMALLINT(__threadSetMaxStackSize(_intVal(i), _intVal(limit))) );
    }
%}
.
    ^ nil
! !

!Process methodsFor:'private scheduler access'!

setPriority:aNumber
    "set priority without telling processor - not for public use"

    prio := aNumber
!

setStartBlock:aBlock
    "set the receivers startup block"

    startBlock := aBlock
!

setStateTo:newState if:oldState
    state == oldState ifTrue:[state := newState]
!

setStateTo:newState if:oldState1 or:oldState2
    (state == oldState1 or:[state == oldState2]) ifTrue:[state := newState]
!

setId:idNumber state:stateSymbol
    "set id and state - not for public use"

    id := idNumber.
    state := stateSymbol.
!

start
    "start the process - this is sent by the VM to the process to get
     the process up and running.
     Sending start to the process allows more flexible handling
     of processes, since anything that responds to #start can be handled 
     by the VM then ..."

    |block|

    (block := startBlock) notNil ifTrue:[
	"
	 just for your convenience ...
	"
	name isNil ifTrue:[
	    name := '(' , startBlock displayString , ')'
	].
	startBlock := nil.
	TerminateSignal handle:[:ex |
	    ex return
	] do:[
	    block value
	].
	(block := exitAction) notNil ifTrue:[
	    exitAction := nil.
	    block value.
	].
	Processor terminateActiveNoSignal
    ] ifFalse:[
	"is this artificial restriction useful ?"
	self error:'a process cannot be started twice'
    ]
! !

!Process methodsFor:'suspend / resume'!

suspend
    "suspend the receiver process - will continue to run when a resume is sent"

    Processor suspend:self
!

resume
    "resume the receiver process"

    Processor resume:self
!

resumeForSingleSend
    "resume the receiver process, but only let it execute a single send."

    Processor resumeForSingleSend:self
!

terminate
    "terminate the receiver process. All unwind actions and the exit-action (if any)
     will be performed before the process is really terminated."

    Processor activeProcess == self ifTrue:[
	Signal noHandlerSignal handle:[:ex |
	    ex return.
	] do:[
	    TerminateSignal raise.
	].
	Processor terminateNoSignal:self
    ] ifFalse:[
	self interruptWith:[self terminate]
    ]
!

terminateNoSignal
    "terminate the receiver process without performing any unwind- or exit-actions"

    Processor terminateNoSignal:self
! !

!Process methodsFor:'interrupts'!

interruptWith:aBlock
    "interrupt the receiver and make it evaluate aBlock.
     If the receiver is currently suspended, the block will be remembered
     to be evaluated once the receiver wakes up."

    self uninterruptablyDo:[
	interruptActions isNil ifTrue:[
	    interruptActions := OrderedCollection with:aBlock.
	] ifFalse:[
	    interruptActions addLast:aBlock.
	].
    ].
    Processor scheduleForInterrupt:self.
!

interrupt
    "evaluate my interrupt-actions
     the process will go back to where it got interrupted
     after doing this.
    "
    |action|

    [interruptActions notNil and:[interruptActions notEmpty]] whileTrue:[
	action := interruptActions removeFirst.
	action value
    ].
    interruptActions := nil
! !

!Process methodsFor:'special'!

withPriority:aPrio do:aBlock
    "execute aBlock at another priority. This can be used to perform
     time-consuming operations at a more user-friendly priority,
     or some critical action at a higher priority. Do not use too high
     of a priority to avoid locking up the system (event processing takes place
     at 24)"

    |oldprio|

    oldprio := prio.
    self priority:aPrio.

    ^ aBlock valueNowOrOnUnwindDo:[
	self priority:oldprio
    ]
!

withLowerPriorityDo:aBlock
    "execute aBlock at a lower priority. This can be used to perform
     time-consuming operations at a more user-friendly priority."

    ^ self withPriority:(prio - 1) do:aBlock
!

uninterruptablyDo:aBlock
    "execute aBlock with interrupts blocked. This does not prevent
     preemption by higher priority processes."

    |wasBlocked|

    "we must keep track of blocking-state if this is called nested"

    wasBlocked := OperatingSystem blockInterrupts.
    ^ aBlock valueNowOrOnUnwindDo:[
	wasBlocked ifFalse:[OperatingSystem unblockInterrupts]
    ]
! !
    
!Process methodsFor:'printing & storing'!

printOn:aStream
    aStream nextPutAll:state article;
	    space;
	    nextPutAll:state;
	    nextPutAll:' Process (';
	    nextPutAll:self nameOrId;
	    nextPutAll:')'
! !
